use std::sync::{Arc, Once};

use axum::{
    body::Body,
    extract::State,
    http::{header, Response, StatusCode, Uri},
    response::{Html, IntoResponse},
    routing::get,
};
use axum_typed_routing::{route, TypedRouter};
use maud::{html, Markup};
use rust_embed::RustEmbed;

#[allow(unused)]
#[route(GET "/item/:id?amount&offset")]
async fn item_handler(
    id: u32,
    amount: Option<u32>,
    offset: Option<u32>,
    State(state): State<AppState>,
    // Json(json): Json<u32>,
) -> Markup {
    // todo!("handle request")

    html! {
        h1 { "Item" }
        p {
            (format!("{id:?} {amount:?} {offset:?} {state:?}"))
        }
    }
}

// We use a wildcard matcher ("/dist/*file") to match against everything
// within our defined assets directory. This is the directory on our Asset
// struct below, where folder = "examples/public/".
#[route(GET "/dist/*path")]
async fn static_handler(path: String, _: State<AppState>) -> impl IntoResponse {
    let path = path.trim_start_matches('/').to_string();

    StaticFile(path)
}

fn markup_404(uri: String) -> Markup {
    html! {
        h1 { "404" }
        p { (format!("{uri:?} Not Found")) }
        @for i in 0..5 {
            div .{"m-" (i)} { (i) }
        }
    }
}

fn markup_405() -> Markup {
    html! {
        h1 { "404" }
        p { "Method not allowed!" }
        @for i in 1..3 {
            div .{"m-" (i)} {}
        }
    }
}

// Finally, we use a fallback route for anything that didn't match.
async fn handle_404(uri: Uri) -> impl IntoResponse {
    (StatusCode::NOT_FOUND, markup_404(format!("{uri:?}"))).into_response()
}

async fn handle_405() -> impl IntoResponse {
    (StatusCode::METHOD_NOT_ALLOWED, markup_405()).into_response()
}

#[derive(RustEmbed)]
#[folder = "public/"]
struct Asset;

pub struct StaticFile<T>(pub T);

impl<T> IntoResponse for StaticFile<T>
where
    T: Into<String>,
{
    fn into_response(self) -> Response<Body> {
        let path = self.0.into();
        tracing::debug!(?path);

        match Asset::get(path.as_str()) {
            Some(content) => {
                let mime = mime_guess::from_path(path).first_or_octet_stream();

                ([(header::CONTENT_TYPE, mime.as_ref())], content.data).into_response()
            }
            None => (StatusCode::NOT_FOUND, markup_404(path)).into_response(),
        }
    }
}

pub fn initialize_logger() {
    static INIT: Once = Once::new();

    INIT.call_once(|| {
        let env_filter = tracing_subscriber::EnvFilter::builder()
            .with_default_directive(tracing::level_filters::LevelFilter::INFO.into())
            .from_env_lossy();

        tracing_subscriber::fmt()
            .pretty()
            .with_env_filter(env_filter)
            .init();
    });
}

use minijinja::{Environment, Template};

#[derive(Clone, Debug)]
struct AppState {
    _field: String,
    mj_env: Arc<minijinja::Environment<'static>>,
}

#[route(GET "/")]
async fn jinja_index_handler(state: State<AppState>) -> impl IntoResponse {
    RenderTemplate::new("/".to_string(), state.mj_env.clone())
}

#[route(GET "/*path")]
async fn jinja_index_handler_path(path: String, state: State<AppState>) -> impl IntoResponse {
    RenderTemplate::new(path, state.mj_env.clone())
}

struct RenderTemplate {
    tmpl_name: String,
    env: Arc<Environment<'static>>,
    ctx: minijinja::Value,
    block: Option<String>,
}

impl RenderTemplate {
    fn new(tmpl_name: String, env: Arc<Environment<'static>>) -> Self {
        Self {
            tmpl_name,
            env,
            ctx: minijinja::Value::from(()),
            block: None,
        }
    }

    fn _new_with_ctx(
        tmpl_name: String,
        env: Arc<Environment<'static>>,
        ctx: minijinja::Value,
    ) -> Self {
        Self {
            tmpl_name,
            env,
            ctx,
            block: None,
        }
    }
}

impl IntoResponse for RenderTemplate {
    fn into_response(self) -> axum::response::Response {
        let path = self.tmpl_name;
        let env = self.env;
        let ctx = self.ctx;
        let block = self.block;

        let res = env.get_template(&path);

        let render = move |template: Template| match block {
            None => match template.render(ctx) {
                Ok(html) => Html(html).into_response(),
                Err(_) => (
                    StatusCode::INTERNAL_SERVER_ERROR,
                    "Failed to render template",
                )
                    .into_response(),
            },
            Some(block) => match template.eval_to_state(ctx).unwrap().render_block(&block) {
                Ok(html) => Html(html).into_response(),
                Err(_) => (
                    StatusCode::INTERNAL_SERVER_ERROR,
                    "Failed to render block template",
                )
                    .into_response(),
            },
        };

        match res {
            Ok(template) => render(template),
            Err(_) => {
                let template = env.get_template("404").unwrap();
                let html = render(template);
                (StatusCode::NOT_FOUND, html).into_response()
            }
        }
    }
}

/// Template loader for embedded templates. Uses Symbol’s value as variable is void: rust-embed to embed templates into the binary.
/// For input path example/template
/// it looks up:
/// - $template_dir/example/template/index.html
/// - $template_dir/example/template.html
///
/// Instead of example/template
/// any of:
/// - example//template/
/// - example/template///
/// - example/template.html
/// - example/template/index.html
/// would also resolve to the same 2 templates above
/// TODO Canonilize path with a middleware that redirects any weird variants to the canonical path?
fn template_loader(name: &str) -> Result<Option<String>, minijinja::Error> {
    #[derive(RustEmbed)]
    #[folder = "src/templates/"]
    struct Templ;
    // lets extract the "clean path" (i.e. '.html' or '/index.html' suffix)
    let mut clean_path = name
        .to_string()
        .chars()
        .fold(String::new(), |mut acc, c| {
            if c == '/' && acc.chars().last() != Some('/') {
                acc.push(c);
            } else if c != '/' {
                acc.push(c);
            }

            acc
        })
        .to_string();
    clean_path = clean_path
        .strip_suffix("index.html")
        .unwrap_or(&clean_path)
        .to_string();

    clean_path = clean_path
        .strip_suffix(".html")
        .unwrap_or(&clean_path)
        .to_string();

    let cleaned_path = clean_path.to_string();

    let files_to_try = vec![
        format!("./{}.html", cleaned_path),
        format!("./{}/index.html", cleaned_path),
    ];

    let found_template = files_to_try.into_iter().find_map(|path| {
        tracing::info!(?path);
        match Templ::get(path.as_str()) {
            Some(content) => {
                let content = std::str::from_utf8(&content.data).unwrap().to_string();
                Some(content)
            }
            None => None,
        }
    });

    Ok(found_template)
}

#[tokio::main]
async fn main() {
    initialize_logger();

    let mut mj_env = Environment::new();

    mj_env.set_loader(template_loader);

    let app_state = AppState {
        _field: "".to_string(),
        mj_env: Arc::new(mj_env),
    };

    // TODO pick free port/config
    let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();

    let ui_router = axum::Router::new().typed_route(item_handler);

    let router: axum::Router = axum::Router::new()
    // .route("/", get(jinja_index_handler))
        .merge(ui_router.clone())
        .nest("/ui", ui_router)
        .typed_route(jinja_index_handler)
        .typed_route(jinja_index_handler_path)
        .typed_route(static_handler)
        .fallback_service(get(handle_404))
        .method_not_allowed_fallback(handle_405)
        .with_state(app_state);

    let _ = axum::serve(listener, router.into_make_service()).await;
}
